/*
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <X11/IntrinsicP.h>     /* Intrinsics header file */
#include <X11/StringDefs.h>     /* Resource string definitions */

#include "SortWdgtP.h"

#include <objc.h>
#include <stdlib.h>
#include <assert.h>

#define BAR_WIDTH 5
#define HORZ_SPACE 2

#define HORZ_MARGIN 2
#define VERT_MARGIN 2
#define BORDER_WIDTH 1

#define Offset(field) XtOffsetOf(SortRec, sort.field)

static XtResource resources[] = {
    {XtNlabel, XtCLabel, XtRString, sizeof(char*), Offset(label), XtRString, (XtPointer) NULL},
    {XtNfont,  XtCFont, XtRFontStruct, sizeof(XFontStruct *), Offset(font), XtRString, (XtPointer) XtDefaultFont},
    {XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel), Offset(foreground), XtRString, (XtPointer) XtDefaultForeground},
};

static void classinitialize(void)
{
  dbg("classinitialize\n");
}

static void classpartinitialize(WidgetClass cls)
{
  dbg("classpartinitialize\n");
}

static GC getnormalgc(SortWidget lw)
{
  XGCValues values;
  values.foreground = lw->sort.foreground;
  values.font = lw->sort.font->fid;
  return XtGetGC((Widget)lw,GCForeground|GCFont,&values);
}

static void initialize(Widget request,Widget new,ArgList args,Cardinal *num)
{
  SortWidget sw = (SortWidget)new;
  dbg("initialize\n");
  sw->sort.gc = getnormalgc(sw);
  sw->sort.count = 0;
  sw->sort.data = NULL;
  sw->sort.label = NULL;
}

static void destroy(Widget w)
{
  SortWidget sw = (SortWidget)w;
  dbg("destroy\n");
  if (sw->sort.data) free(sw->sort.data);
  if (sw->sort.label) XtFree((char *)sw->sort.label);
  XtReleaseGC(w,sw->sort.gc);
}

static void showbar(SortWidget sw,int i)
{
  int x,y;
  GC gc = sw->sort.gc;
  int w = BAR_WIDTH;
  int h = sw->sort.data[i];
  int fonth = sw->sort.font->max_bounds.ascent; 

  x = HORZ_MARGIN + i * (HORZ_SPACE + BAR_WIDTH);
  y = VERT_MARGIN + fonth + sw->sort.maxval - h;
  XFillRectangle(XtDisplay((Widget)sw),XtWindow((Widget)sw),gc,x,y,w,h);
}

static void clearbar(SortWidget sw,int i)
{
  int x,y;
  int w = BAR_WIDTH;
  int h = sw->sort.maxval; /* clear maxval */
  int fonth = sw->sort.font->max_bounds.ascent; 

  x = HORZ_MARGIN + i * (HORZ_SPACE + BAR_WIDTH);
  y = VERT_MARGIN + fonth;
  XClearArea(XtDisplay((Widget)sw),XtWindow((Widget)sw),x,y,w,h,TRUE);
}

static void redisplay(Widget w,XEvent *event,Region region)
{
  char*msg;
  SortWidget sw = (SortWidget)w;

  dbg("redisplay\n");

  if ((msg = sw->sort.label)) {
    GC gc = sw->sort.gc;
    int x = HORZ_MARGIN;
    int y = VERT_MARGIN + sw->sort.font->max_bounds.ascent;
    XDrawString(XtDisplay(w),XtWindow(w),gc,x,y,msg,strlen(msg));
  }

  if (sw->sort.data) {
    int i,n = sw->sort.count;
    for(i=0;i<n;i++) showbar(sw,i); 
  }
}

static void resize(Widget w)
{
  dbg("resize\n");
  if (XtIsRealized(w)) {
    XClearWindow(XtDisplay(w),XtWindow(w));
    (*(XtClass(w)->core_class.expose))(w,NULL,NULL);
  }
}

static XtGeometryResult querygeometry(Widget w,XtWidgetGeometry *proposed,XtWidgetGeometry *desired)
{
  dbg("querygeometry\n");
  return XtGeometryAlmost;
}

void setbardata(Widget w,int *e,int n,int maxval)
{
  int i;
  SortWidget sw;
  XFontStruct *fs;
  dbg("setbardata\n");
  XtCheckSubclass(w,sortWidgetClass, NULL);
  sw = (SortWidget)w;
  if (n) {
    sw->core.width=2 * HORZ_MARGIN + n * BAR_WIDTH + (n-1) * HORZ_SPACE;
    sw->core.height=2 * VERT_MARGIN + maxval;
  }
  if ((fs=sw->sort.font)) {
    sw->core.height+=fs->max_bounds.ascent+fs->max_bounds.descent;
  }
  if (sw->sort.data) free(sw->sort.data);
  sw->sort.count = n;
  sw->sort.maxval = maxval;
  sw->sort.data  = (int*)malloc(sizeof(int)*n);
  for(i=0;i<n;i++) sw->sort.data[i] = e[i];
  resize(w);
}

void setbarlabel(Widget w,const char *msg)
{
  SortWidget sw;
  dbg("setbarlabel\n");
  XtCheckSubclass(w,sortWidgetClass, NULL);
  sw = (SortWidget)w;
  if (sw->sort.label) XtFree((char *)sw->sort.label);
  sw->sort.label = XtNewString((char*)msg);
  resize(w);
}

void swapbars(Widget w,int i,int a,int j,int b)
{
  SortWidget sw;
  dbg("swapbars %i %i %i %i\n",i,a,j,b);
  XtCheckSubclass(w,sortWidgetClass, NULL);
  sw = (SortWidget)w;
  assert((i < sw->sort.count) && (j < sw->sort.count));
  sw->sort.data[i] = b;
  sw->sort.data[j] = a;

  clearbar(sw,i);
  clearbar(sw,j);
}

void movebar(Widget w,int a,int j,int b)
{
  SortWidget sw;
  dbg("movebar %i %i %i\n",a,j,b);
  XtCheckSubclass(w,sortWidgetClass, NULL);
  sw = (SortWidget)w;

  assert(j < sw->sort.count);
  sw->sort.data[j] = a;

  clearbar(sw,j);
}

void comparebars(Widget w,int i,int a,int j,int b)
{
  SortWidget sw;
  dbg("comparebars %i %i %i %i\n",i,a,j,b);
  XtCheckSubclass(w,sortWidgetClass, NULL);
  sw = (SortWidget)w;
}

/* Class record declaration */

SortClassRec sortClassRec = {
 /* Core class part */
 {
    /* superclass            */ (WidgetClass) &widgetClassRec,
    /* class_name            */ "Sort",
    /* widget_size           */ sizeof(SortRec),
    /* class_initialize      */ classinitialize,
    /* class_part_initialize */ classpartinitialize,
    /* class_inited          */ FALSE,
    /* initialize            */ initialize,
    /* initialize_hook       */ NULL,
    /* realize               */ XtInheritRealize,
    /* actions               */ NULL,
    /* num_actions           */ 0,
    /* resources             */ resources,
    /* num_resources         */ XtNumber(resources),
    /* xrm_class             */ NULLQUARK,
    /* compress_motion       */ TRUE,
    /* compress_exposure     */ XtExposeCompressMultiple,
    /* compress_enterleave   */ TRUE,
    /* visible_interest      */ FALSE,
    /* destroy               */ destroy,
    /* resize                */ resize,
    /* expose                */ redisplay,
    /* set_values            */ NULL,
    /* set_values_hook       */ NULL,
    /* set_values_almost     */ XtInheritSetValuesAlmost,
    /* get_values_hook       */ NULL,
    /* accept_focus          */ NULL,
    /* version               */ XtVersion,
    /* callback offsets      */ NULL,
    /* tm_table              */ NULL,
    /* query_geometry        */ querygeometry,
    /* display_accelerator   */ NULL,
    /* extension             */ NULL
  },
};

WidgetClass sortWidgetClass = (WidgetClass) &sortClassRec;

